import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Input, Embedding, Dot, Reshape
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Lambda

import matplotlib
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

import matplotlib.font_manager as fm
from sklearn.metrics.pairwise import cosine_similarity
from collections import Counter
import random
import math


corpus = [line.strip() for line in open('politician_text_tokens.txt', encoding='utf8').readlines()]
tweet_labels = []
tweet_tags = []
for line in [line.strip() for line in open('politician_text_tokens_label_tag.txt', encoding='utf8').readlines()]:
    tks = line.split('\t')
    if len(tks) < 2:
        tweet_labels.append(int(tks[0]))
        tweet_tags.append('')
    else:
        tweet_labels.append(int(tks[0]))
        tweet_tags.append(tks[1].split(','))

word_to_idx = np.load('word_to_embedding_idx.npy', allow_pickle=True).item()
D = len(word_to_idx)+1
index_to_word = {idx : word for word, idx in word_to_idx.items()}

def corpus_to_train(text):
    cooc = []
    voca = set()
    tks = [w for w in text.split(' ') if w in word_to_idx]
    voca.update(tks)
    if len(voca) == 0:
        return None, list()
    for i in range(len(tks)-1):
        for j in range(i+1, len(tks)):
            cooc.append((word_to_idx[tks[i]], word_to_idx[tks[j]]))
    cooc_count = Counter(cooc)
    
    neg_voca = word_to_idx.keys() - voca
    voca = list(voca)
    neg_voca = list(neg_voca)
    
    data = []
    for head,tail in cooc_count:
        data.append((head, tail, cooc_count[(head,tail)]))
        
    for word in voca:
        negs = random.choices(neg_voca, k=10)
        for neg in negs:
            data.append((word_to_idx[word], word_to_idx[neg], 0))
    data = np.array(data)
    return data[:,:2], data[:,2]

# Build the model
class AttLayer(tf.keras.layers.Layer):
    def __init__(self, units, **kwargs):
        super(AttLayer, self).__init__(**kwargs)
        self.units = units

    def build(self, input_shape):
        # Create a trainable weight variable (kernel)
        self.kernel = self.add_weight("kernel", shape=(self.units, self.units), trainable=True)

    def call(self, inputs):
        # Apply the kernel to the inputs
        return tf.matmul(inputs, self.kernel)

embedding_dim = 21
input_head = tf.keras.layers.Input((1,))
input_tail = tf.keras.layers.Input((1,))
input_start_idx = tf.keras.layers.Input(())
input_end_idx = tf.keras.layers.Input(())

embedding_layer = Embedding(D, embedding_dim, input_length=1)
head_embedding = embedding_layer(input_head)
tail_embedding = embedding_layer(input_tail)

AttK = AttLayer(embedding_dim-1)(embedding_layer(input_head)[:,:,:20])
AttQ = AttLayer(embedding_dim-1)(embedding_layer(input_head)[:,:,:20])

dot_product = tf.keras.layers.Dot(axes=2)([head_embedding, tail_embedding])
dot_product = tf.keras.layers.Reshape((1,))(dot_product)

model = Model(inputs=[input_head, input_tail], outputs=[dot_product, embedding_layer(input_head), AttK, AttQ])

optimizer = tf.keras.optimizers.Adam()

loss_1 = tf.keras.losses.MeanSquaredError()

def irm_penalty(pred, labels):
    with tf.GradientTape() as penalty_tape:
        penalty_tape.watch(pred)
        loss = tf.reduce_mean(loss_1(labels, pred))
    grads = penalty_tape.gradient(loss, pred)
    return tf.reduce_mean(tf.square(grads))

# Batch training
batch_size = 16
epochs = 10

irm_lambda = 10
for epoch in range(epochs):
    num_batches = len(corpus) // batch_size
    loss_total, loss_total1, loss_total2, loss_total3 = 0,0,0,0
    
    for batch_idx in range(num_batches):
        start_idx = batch_idx * batch_size
        end_idx = (batch_idx + 1) * batch_size
        batch_pairs = []
        batch_labels = []
        batch_tweet_labels = []
        batch_tag_mask = []
        batch_tweet_number = []
        for i in range(start_idx, end_idx):
            # Convert skipgram sequences to NumPy arrays
            pairs, labels = corpus_to_train(corpus[i])
            if len(labels) == 0:
                continue
        
            batch_pairs.extend(pairs)
            batch_labels.extend(labels)
            batch_tweet_labels.extend([tweet_labels[i]] * len(labels))

            tags = tweet_tags[i]
            tag_idx = [word_to_idx[tag] for tag in tags if tag in word_to_idx]
            for j, (w,c) in enumerate(pairs):
                if labels[j] == 0:
                    batch_tag_mask.append(0)
                    batch_tweet_number.append(-1)
                    continue
                    
                if w in tag_idx:
                    batch_tag_mask.append(1)
                else:
                    batch_tag_mask.append(0)
                batch_tweet_number.append(i)

        batch_pairs = np.array(batch_pairs, dtype=np.int32)
        batch_labels = np.array(batch_labels, dtype=np.int32)
        
        with tf.GradientTape() as tape:
    
            predictions, embs, AttK, AttQ = model([np.expand_dims(batch_pairs[:, 0], axis=1), np.expand_dims(batch_pairs[:, 1], axis=1)], training=True)
            AttK = tf.squeeze(AttK)
            AttQ = tf.squeeze(AttQ)
            loss2 = 0
            loss2_N = 0
            
            for i in range(start_idx, end_idx):
                m = 0
                sub_idx = []
                for j in range(embs.shape[0]):
                    if not batch_tweet_number[j] == i:
                        continue
                    if batch_tag_mask[j] == 0:
                        continue
                    sub_idx.append(j)
                    m += 1

                if m > 0:
                    AttK = tf.gather(AttK, sub_idx)
                    AttQ = tf.gather(AttQ, sub_idx)
                    
                    AttAlpha = tf.nn.softmax(tf.matmul(tf.matmul(AttK, tf.transpose(AttQ))/math.sqrt(m), tf.expand_dims(tf.ones(m), axis=-1)))

                    ztp = tf.reduce_sum(AttAlpha * tf.gather(embs[:,0,20], sub_idx))

                    loss2 += max(0, 1-tweet_labels[i]*ztp)
                    loss2_N += 1
            if loss2 == 0:
                continue                
            else:
                loss2 = loss2/loss2_N
                    
            irm_e1 = 0
            irm_e2 = 0
            e1_selector = [v == -1 for v in batch_tweet_labels]
            e1_pred = tf.boolean_mask(predictions, e1_selector)
            e1_label = tf.boolean_mask(batch_labels, e1_selector)
            e2_selector = [v == 1 for v in batch_tweet_labels]
            e2_pred = tf.boolean_mask(predictions, e2_selector)
            e2_label = tf.boolean_mask(batch_labels, e2_selector)
            
            if len(e1_pred) > 0:
                irm_e1 += irm_penalty(tf.squeeze(tf.stack(e1_pred)), e1_label)
            if len(e2_pred) > 0:
                irm_e2 += irm_penalty(tf.squeeze(tf.stack(e2_pred)), e2_label)
            irm_loss = (irm_e1 + irm_e2)/2
            
            loss1 = tf.reduce_mean(loss_1(np.expand_dims(batch_labels, axis=1), predictions))
            loss_all = loss1+loss2+irm_lambda*irm_loss
            
            
        gradients = tape.gradient(loss_all, model.trainable_variables)
        optimizer.apply_gradients(zip(gradients, model.trainable_variables))
        if (batch_idx % 1000) == 0:
            print('batch:{} loss1:{} loss2:{} irm_loss:{}'.format(batch_idx, loss1, loss2, irm_loss))
        
        loss_total += loss_all.numpy()
        loss_total1 += loss1.numpy()
        loss_total2 += loss2
        loss_total3 += irm_loss
        
    print(f'Epoch {epoch+1}, Loss: {loss_total / num_batches} Loss1: {loss_total1 / num_batches} Loss2: {loss_total2 / num_batches} IRM Loss:{loss_total3 / num_batches}')


# Get word embeddings
word_embeddings = embedding_layer.get_weights()[0]

# Save the embeddings
np.save('word_embeddings_BIDEL.npy', word_embeddings)

# Load the word embeddings
# word_embeddings = np.load('word_embeddings.npy')
word_embeddings = embedding_layer.get_weights()[0]

zp = word_embeddings[:,20]

with open('word_political_bias.txt', 'w', encoding='utf8') as file:
    for i in range(1,len(zp)-1):
        file.write('{}\t{}\n'.format(index_to_word[i], zp[i]))


# visualization of a selection of words
plt.rcParams['font.family'] = 'MS Gothic'
words_to_visualize = [line.strip() for line in open('vis_hashtags_small.txt', encoding='utf8')]

selected_word_indices = [word_to_idx[word] for word in words_to_visualize]
selected_word_embeddings = word_embeddings[selected_word_indices]

zs = selected_word_embeddings[:,:20]
zp = selected_word_embeddings[:,20]

print(words_to_visualize)
print(zp)

# perform dimensionality reduction using t-SNE
tsne = TSNE(n_components=1, perplexity=10, n_iter=300)
tsne_result = tsne.fit_transform(zs)

result = tsne_result
# Plot the word embeddings
plt.figure(figsize=(15, 12))
plt.scatter(zp, result[:, 0], marker='o')

for i, word in enumerate(words_to_visualize):
    plt.annotate(word, xy=(zp[i], result[i, 0]), fontsize=12)

plt.xlabel('Polarity')
plt.ylabel('Semantics')
plt.title('Word Embeddings Visualization')
plt.show()
